"""
Created on 5 Apr 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import json

from http import HTTPStatus

from scs_core.aws.security.cognito_device import CognitoDeviceIdentity

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class CognitoDeviceFinder(object):
    """
    classdocs
    """

    __URL = 'https://6c2sfqt656.execute-api.us-west-2.amazonaws.com/default/CognitoDevices/retrieve'

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, http_client):
        self.__http_client = http_client                    # requests package
        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    def find_all(self, token):
        url = '/'.join((self.__URL, 'all'))

        response = self.__http_client.get(url, headers=self.__headers(token))
        self.__check_response(response)

        # print("response: %s" % response.json())

        return (CognitoDeviceIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def find_by_tag(self, token, tag):
        url = '/'.join((self.__URL, 'in'))
        payload = json.dumps({"username": tag})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return (CognitoDeviceIdentity.construct_from_jdict(jdict) for jdict in response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def __headers(self, token):
        headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/json", "Token": token}
        self.__logger.debug('headers: %s' % headers)

        return headers


    def __check_response(self, response):
        self.__logger.debug('response: %s' % response.json())

        status = HTTPStatus(response.status_code)

        if status != HTTPStatus.OK:
            raise HTTPException.construct(status.value, response.reason, response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoDeviceFinder:{http_client:%s}" % self.__http_client



"""
Created on 2 April 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://en.wikipedia.org/wiki/List_of_HTTP_status_codes
"""


# TODO: remove, and replace with httpstatus package
# --------------------------------------------------------------------------------------------------------------------

class HTTPStatus(object):
    """
    classdocs
    """

    OK =                200
    CREATED =           201
    NO_CONTENT =        204

    BAD_REQUEST =       400
    FORBIDDEN =         403
    NOT_FOUND =         404



"""
Created on 22 Nov 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://docs.aws.amazon.com/cognito/latest/developerguide/signing-up-users-in-your-app.html
https://docs.aws.amazon.com/cognito/latest/developerguide/user-pool-settings-policies.html
https://stackoverflow.com/questions/2520893/how-to-flush-the-input-stream-in-python

example document (credentials):
{"email": "bruno.beloff@southcoastscience.com", "password": "pass"}

example document (identity):
{"username": "8", "creation-date": "2021-11-24T12:51:12Z", "confirmation-status": "CONFIRMED", "enabled": true,
"email": "bruno.beloff@southcoastscience.com", "given-name": "Bruno", "family-name": "Beloff", "is-super": true}
"""

import ast
import json
import re
import sys
import termios

from collections import OrderedDict
from getpass import getpass

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.json import JSONable, MultiPersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class CognitoUserCredentials(MultiPersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "cognito_user_credentials.json"

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def from_stdin(cls, name):
        line = sys.stdin.readline()
        jdict = json.loads(line)

        return cls.construct_from_jdict(jdict, name=name)


    @classmethod
    def from_user(cls, name):
        try:
            termios.tcflush(sys.stdin, termios.TCIOFLUSH)               # flush stdin
        except termios.error:
            pass

        print("Enter email address: ", end="", file=sys.stderr)
        email = input().strip()

        print("Enter password: ", end="", file=sys.stderr)
        password = input().strip()

        return cls(name, email, password)


    @staticmethod
    def password_from_user():
        try:
            termios.tcflush(sys.stdin, termios.TCIOFLUSH)               # flush stdin
        except termios.error:
            pass

        return getpass().strip()


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def persistence_location(cls, name):
        filename = cls.__FILENAME if name is None else '_'.join((name, cls.__FILENAME))

        return cls.aws_dir(), filename


    @classmethod
    def construct_from_jdict(cls, jdict, name=None, skeleton=False):
        if not jdict:
            return cls(None, None, None) if skeleton else None

        email = jdict.get('email')
        password = jdict.get('password')

        return cls(name, email, password)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name, email, password):
        """
        Constructor
        """
        super().__init__(name)

        self.__email = email                            # string
        self.__password = password                      # string


    # ----------------------------------------------------------------------------------------------------------------

    def ok(self):
        if not self.email or not self.password:
            return False

        if not Datum.is_email_address(self.email):
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['email'] = self.email
        jdict['password'] = self.password

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def email(self):
        return self.__email


    @email.setter
    def email(self, email):
        self.__email = email


    @property
    def password(self):
        return self.__password


    @password.setter
    def password(self, password):
        self.__password = password


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoUserCredentials:{name:%s, email:%s, password:%s}" %  \
               (self.name, self.email, self.password)


# --------------------------------------------------------------------------------------------------------------------

class CognitoUserIdentity(JSONable):
    """
    classdocs
    """

    __STATUSES = {
        'U': 'UNCONFIRMED',
        'C': 'CONFIRMED',
        'P': 'PASSWORD_RESET_REQUIRED',
        'F': 'FORCE_CHANGE_PASSWORD',
        'D': 'DISABLED'
    }

    @classmethod
    def status_codes(cls):
        return cls.__STATUSES.keys()


    @classmethod
    def status(cls, code):
        return cls.__STATUSES[code]


    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def is_valid_password(password):
        if not isinstance(password, str):
            return False

        if not 7 < len(password) < 99:
            return False

        if not re.findall(r'[0-9]', password):
            return False

        if not re.findall(r'[A-Z]', password):
            return False

        if not re.findall(r'[a-z]', password):
            return False

        if not re.findall(r'[\^$*.\[\]{}()?"!@#%&/\\,><\':;|_~`]', password):
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_response(cls, res, multiples=False):
        if not res:
            return None

        username = res['Username']

        final_d = {}
        attrs = res['Attributes'] if multiples else res['UserAttributes']

        for attr in attrs:
            nk = None
            for value in attr.values():
                if nk:
                    final_d[nk] = value
                    break

                nk = value

        creation_date = LocalizedDatetime.construct_from_aws(str(res["UserCreateDate"]))
        confirmation_status = res["UserStatus"]
        enabled = res["Enabled"]

        try:
            return cls(username, creation_date, confirmation_status, enabled, final_d['email'], final_d['given_name'],
                       final_d['family_name'], None, ast.literal_eval(final_d['custom:super']))
        except KeyError:
            return None


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls(None, None, None, None, None, None, None, None) if skeleton else None

        username = jdict.get('username')
        creation_date = LocalizedDatetime.construct_from_iso8601(jdict.get('creation_date'))
        confirmation_status = jdict.get('confirmation_status')
        enabled = jdict.get('enabled')
        email = jdict.get('email')
        given_name = jdict.get('given_name')
        family_name = jdict.get('family_name')
        password = jdict.get('password')
        is_super = jdict.get('is_super')

        return cls(username, creation_date, confirmation_status, enabled,
                   email, given_name, family_name, password, is_super=is_super)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, username, creation_date, confirmation_status, enabled,
                 email, given_name, family_name, password, is_super=False):
        """
        Constructor
        """
        self.__username = username                              # string (int)
        self.__creation_date = creation_date                    # LocalisedDatetime
        self.__confirmation_status = confirmation_status        # string
        self.__enabled = enabled                                # bool or None
        self.__email = email                                    # string
        self.__given_name = given_name                          # string
        self.__family_name = family_name                        # string
        self.__password = password                              # string
        self.__is_super = bool(is_super)                        # bool


    def __lt__(self, other):
        if self.family_name is not None:
            if self.family_name.lower() < other.family_name.lower():
                return True

            if self.family_name.lower() > other.family_name.lower():
                return False

        if self.given_name is not None:
            if self.given_name.lower() < other.given_name.lower():
                return True

            if self.given_name.lower() > other.given_name.lower():
                return False

        if self.email.lower() < other.email.lower():
            return True

        if self.email.lower() > other.email.lower():
            return False

        return False


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.username is not None:
            jdict['username'] = self.username

        if self.creation_date is not None:
            jdict['creation_date'] = self.creation_date.as_iso8601()

        if self.confirmation_status is not None:
            jdict['confirmation_status'] = self.confirmation_status

        if self.enabled is not None:
            jdict['enabled'] = self.enabled

        jdict['email'] = self.email

        if self.password is not None:
            jdict['password'] = self.password

        jdict['given_name'] = self.given_name
        jdict['family_name'] = self.family_name
        jdict['is_super'] = self.is_super

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def username(self):
        return self.__username


    @property
    def creation_date(self):
        return self.__creation_date


    @property
    def confirmation_status(self):
        return self.__confirmation_status


    @property
    def enabled(self):
        return self.__enabled


    @property
    def email(self):
        return self.__email


    @property
    def given_name(self):
        return self.__given_name


    @property
    def family_name(self):
        return self.__family_name


    @property
    def password(self):
        return self.__password


    @property
    def is_super(self):
        return self.__is_super


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoUserIdentity:{username:%s, creation_date:%s, confirmation_status:%s, enabled:%s, " \
               "email:%s, given_name:%s, family_name:%s, is_super:%s}" % \
               (self.username, self.creation_date, self.confirmation_status, self.enabled,
                self.email, self.given_name, self.family_name, self.is_super)



"""
Created on 24 Nov 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from http import HTTPStatus

from scs_core.aws.security.cognito_user import CognitoUserIdentity

from scs_core.data.json import JSONify

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class CognitoUserFinder(object):
    """
    classdocs
    """

    __URL = 'https://iw0jza59y1.execute-api.us-west-2.amazonaws.com/default/CognitoUsersFinder/'

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, http_client):
        self.__http_client = http_client                    # requests package
        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    def find_all(self, token):
        url = '/'.join((self.__URL, 'all'))

        response = self.__http_client.get(url, headers=self.__headers(token))
        self.__check_response(response)

        # print("response: %s" % response.json())

        return (CognitoUserIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def find_by_status(self, token, confirmation_status):
        url = '/'.join((self.__URL, confirmation_status.lower()))

        response = self.__http_client.get(url, headers=self.__headers(token))
        self.__check_response(response)

        return (CognitoUserIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def find_by_enabled(self, token, enabled):
        url = '/'.join((self.__URL, 'enabled' if enabled else 'disabled'))

        response = self.__http_client.get(url, headers=self.__headers(token))
        self.__check_response(response)

        return (CognitoUserIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def find_by_email(self, token, email):
        url = '/'.join((self.__URL, 'in'))
        payload = JSONify.dumps({"Email": email})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return (CognitoUserIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def find_by_usernames(self, token, usernames):
        url = '/'.join((self.__URL, 'usernames'))
        payload = JSONify.dumps(usernames)

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return (CognitoUserIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def get_by_email(self, token, email):
        url = '/'.join((self.__URL, 'exact'))
        payload = JSONify.dumps({"Email": email})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return CognitoUserIdentity.construct_from_jdict(response.json())


    def get_self(self, token):
        url = '/'.join((self.__URL, 'self'))

        response = self.__http_client.get(url, headers=self.__headers(token))
        self.__check_response(response)

        return CognitoUserIdentity.construct_from_jdict(response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def __headers(self, token):
        headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/json", "Token": token}
        self.__logger.debug('headers: %s' % headers)

        return headers


    def __check_response(self, response):
        self.__logger.debug('response: %s' % response.json())

        status = HTTPStatus(response.status_code)

        if status != HTTPStatus.OK:
            raise HTTPException.construct(status.value, response.reason, response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoUserFinder:{http_client:%s}" % self.__http_client



"""
Created on 24 Nov 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from http import HTTPStatus

from scs_core.aws.security.cognito_user import CognitoUserIdentity
from scs_core.data.json import JSONify
from scs_core.sys.http_exception import HTTPException


# --------------------------------------------------------------------------------------------------------------------

class CognitoUserCreator(object):
    """
    classdocs
    """

    __AUTHORIZATION = '@southcoastscience.com'
    __URL = 'https://0knr39qhv7.execute-api.us-west-2.amazonaws.com/default/CognitoUserAccountCreator'

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, http_client):
        self.__http_client = http_client                # requests package


    # ----------------------------------------------------------------------------------------------------------------

    def create(self, identity):
        headers = {'Authorization': self.__AUTHORIZATION}

        response = self.__http_client.post(self.__URL, headers=headers, json=identity.as_json())
        status = HTTPStatus(response.status_code)

        if status != HTTPStatus.OK:
            raise HTTPException.construct(status.value, response.reason, response.json())

        return CognitoUserIdentity.construct_from_jdict(response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoUserCreator:{}"


# --------------------------------------------------------------------------------------------------------------------

class CognitoUserEditor(object):
    """
    classdocs
    """

    __URL = 'https://fru3uy2z82.execute-api.us-west-2.amazonaws.com/default/CognitoUserAccounts/edit'

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, http_client, id_token):
        self.__http_client = http_client                    # requests package
        self.__id_token = id_token                          # string


    # ----------------------------------------------------------------------------------------------------------------

    def update(self, identity):
        headers = {'Token': self.__id_token}

        response = self.__http_client.patch(self.__URL, headers=headers, data=JSONify.dumps(identity))
        status = HTTPStatus(response.status_code)

        # print("status_code: %s" % response.status_code)
        # print("text: %s" % response.text)

        if status != HTTPStatus.OK:
            raise HTTPException.construct(status.value, response.reason, response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoUserEditor:{id_token:%s}" % self.__id_token


# --------------------------------------------------------------------------------------------------------------------

class CognitoUserDeleter(object):
    """
    classdocs
    """

    __URL = 'https://fru3uy2z82.execute-api.us-west-2.amazonaws.com/default/CognitoUserAccounts/delete/'

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, http_client, id_token):
        self.__http_client = http_client                    # requests package
        self.__id_token = id_token                          # string


    # ----------------------------------------------------------------------------------------------------------------

    def delete(self, email):
        url = '/'.join((self.__URL, email))
        headers = {'Token': self.__id_token}

        response = self.__http_client.delete(url, headers=headers)
        status = HTTPStatus(response.status_code)

        # print("status_code: %s" % response.status_code)
        # print("text: %s" % response.text)

        if status != HTTPStatus.OK:
            raise HTTPException.construct(status.value, response.reason, response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoUserDeleter:{id_token:%s}" % self.__id_token



"""
Created on 26 Nov 2018

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

This version of MessageManager for AWS lambda function with multi-part response.

Equivalent to cURL:
curl "https://aws.southcoastscience.com/topicMessages?topic=unep/ethiopia/loc/1/climate
&startTime=2018-12-13T07:03:59.712Z&endTime=2018-12-13T15:10:59.712Z"

Test endpoint:
curl "https://hb7aqje541.execute-api.us-west-2.amazonaws.com/default/AWSAggregateTest?topic=unep/ethiopia/loc/1/climate
&startTime=2018-12-13T07:03:59.712Z&endTime=2018-12-13T15:10:59.712Z"
"""

from collections import OrderedDict
from urllib.parse import urlparse, parse_qs

from scs_core.aws.client.rest_client import RESTClient
from scs_core.aws.data.message import Message

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.str import Str
from scs_core.data.timedelta import Timedelta

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class MessageManager(object):
    """
    classdocs AWSAggregateTest
    """

    __REQUEST_PATH = '/topicMessages'

    # __REQUEST_PATH = '/default/AWSAggregateTest'

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, api_key, reporter=None):
        """
        Constructor
        """
        self.__rest_client = RESTClient(api_key)
        self.__reporter = reporter

        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    def find_latest_for_topic(self, topic, end, path, include_wrapper, rec_only, backoff_limit):
        documents = list(self.find_for_topic(topic, None, end, path, False, None, include_wrapper, rec_only, False,
                                             False, True, backoff_limit))
        if documents:
            return documents[0]

        return None


    def find_for_topic(self, topic, start, end, path, fetch_last, checkpoint, include_wrapper, rec_only,
                       min_max, exclude_remainder, fetch_last_written_before, backoff_limit):
        self.__reporter.reset()

        request = MessageRequest(topic, start, end, path, fetch_last, checkpoint, include_wrapper, rec_only,
                                 min_max, exclude_remainder, fetch_last_written_before, backoff_limit)
        self.__logger.debug(request)

        params = request.params()

        # request...
        self.__rest_client.connect()

        try:
            while True:
                jdict = self.__rest_client.get(self.__REQUEST_PATH, params)

                # messages...
                block = MessageResponse.construct_from_jdict(jdict)
                self.__logger.debug(block)

                for item in block.items:
                    yield item

                # report...
                if self.__reporter:
                    self.__reporter.print(len(block), block_start=block.start())

                # next request...
                if block.next_url is None:
                    break

                next_url = urlparse(block.next_url)
                next_params = parse_qs(next_url.query)

                # noinspection PyTypeChecker
                params[MessageRequest.START] = next_params[MessageRequest.START][0]

        finally:
            self.__rest_client.close()


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MessageManager:{rest_client:%s, reporter:%s}" % (self.__rest_client, self.__reporter)


# --------------------------------------------------------------------------------------------------------------------

class MessageRequest(object):
    """
    classdocs
    """

    TOPIC = 'topic'
    START = 'startTime'
    END = 'endTime'
    PATH = 'path'
    FETCH_LAST_WRITTEN = 'fetchLastWritten'
    CHECKPOINT = 'checkpoint'
    INCLUDE_WRAPPER = 'includeWrapper'
    REC_ONLY = 'recOnly'
    MIN_MAX = 'minMax'
    EXCLUDE_REMAINDER = 'excludeRemainder'
    FETCH_LAST_WRITTEN_BEFORE = 'fetchLastWrittenBefore'
    BACKOFF_LIMIT = "backoffLimit"

    # ----------------------------------------------------------------------------------------------------------------

    __DAY = Timedelta(days=1)
    __WEEK = Timedelta(weeks=1)
    __MONTH = Timedelta(days=28)
    __YEAR = Timedelta(days=365)

    @classmethod
    def checkpoint_table(cls, start, end):
        delta = end - start

        if delta < cls.__DAY:
            return None  # raw data rate

        if delta < cls.__WEEK:
            return '**:/01:00'

        if delta < cls.__MONTH:
            return '**:/15:00'

        if delta < cls.__YEAR:
            return '**:00:00'

        return '/06:00:00'


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_qsp(cls, qsp):
        if not qsp:
            return None

        # compulsory...
        topic = qsp.get(cls.TOPIC)
        start = LocalizedDatetime.construct_from_iso8601(qsp.get(cls.START))
        end = LocalizedDatetime.construct_from_iso8601(qsp.get(cls.END))

        # optional...
        path = qsp.get(cls.PATH)
        fetch_last_written = qsp.get(cls.FETCH_LAST_WRITTEN, 'false').lower() == 'true'
        checkpoint = qsp.get(cls.CHECKPOINT)
        include_wrapper = qsp.get(cls.INCLUDE_WRAPPER, 'false').lower() == 'true'
        rec_only = qsp.get(cls.REC_ONLY, 'false').lower() == 'true'
        min_max = qsp.get(cls.MIN_MAX, 'false').lower() == 'true'
        exclude_remainder = qsp.get(cls.EXCLUDE_REMAINDER, 'false').lower() == 'true'
        fetch_last_written_before = qsp.get(cls.FETCH_LAST_WRITTEN_BEFORE, 'false').lower() == 'true'
        backoff_limit = qsp.get(cls.backoff_limit)

        if checkpoint and checkpoint.lower() == 'none':
            checkpoint = None

        return cls(topic, start, end, path, fetch_last_written, checkpoint, include_wrapper, rec_only,
                   min_max, exclude_remainder, fetch_last_written_before, backoff_limit)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, topic, start, end, path, fetch_last_written, checkpoint, include_wrapper, rec_only,
                 min_max, exclude_remainder, fetch_last_written_before, backoff_limit):
        """
        Constructor
        """
        self.__topic = topic  # string
        self.__start = start  # LocalizedDatetime
        self.__end = end  # LocalizedDatetime

        self.__path = path  # string
        self.__fetch_last_written = bool(fetch_last_written)  # bool
        self.__checkpoint = checkpoint  # string
        self.__include_wrapper = bool(include_wrapper)  # bool
        self.__rec_only = bool(rec_only)  # bool
        self.__min_max = bool(min_max)  # bool
        self.__exclude_remainder = bool(exclude_remainder)  # bool
        self.__fetch_last_written_before = bool(fetch_last_written_before)  # bool
        self.__backoff_limit = backoff_limit  # int seconds


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.topic is None or self.start is None or self.end is None:
            return False

        now = LocalizedDatetime.now()

        if self.start > now or self.end > now:
            return False

        if self.start > self.end:
            return False

        if self.include_wrapper and self.checkpoint is not None:
            return False

        if self.rec_only and self.fetch_last_written:
            return False

        if self.rec_only and self.path:
            return False

        if self.rec_only and self.min_max:
            return False

        if self.min_max and self.checkpoint is None:
            return False

        if self.exclude_remainder and self.checkpoint is None:
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    def next_params(self, start):
        return MessageRequest(self.topic, start, self.end, self.path, self.fetch_last_written, self.checkpoint,
                              self.include_wrapper, self.rec_only, self.min_max, self.exclude_remainder,
                              self.backoff_limit, self.fetch_last_written_before).params()


    def change_params(self, start, end):
        return MessageRequest(self.topic, start, end, self.path, self.fetch_last_written, self.checkpoint,
                              self.include_wrapper, self.rec_only, self.min_max, self.exclude_remainder,
                              self.backoff_limit, self.fetch_last_written_before)


    def params(self):
        params = {
            self.TOPIC: self.topic,
            self.START: None if self.start is None else self.start.utc().as_iso8601(include_millis=True),
            self.END: self.end.utc().as_iso8601(include_millis=True)
        }

        if self.path is not None:
            params[self.PATH] = self.path

        if self.fetch_last_written:
            params[self.FETCH_LAST_WRITTEN] = 'true'

        if self.checkpoint is not None:
            params[self.CHECKPOINT] = self.checkpoint

        if self.include_wrapper:
            params[self.INCLUDE_WRAPPER] = 'true'

        if self.rec_only:
            params[self.REC_ONLY] = 'true'

        if self.min_max:
            params[self.MIN_MAX] = 'true'

        if self.exclude_remainder:
            params[self.EXCLUDE_REMAINDER] = 'true'

        if self.fetch_last_written_before:
            params[self.FETCH_LAST_WRITTEN_BEFORE] = 'true'

        if self.backoff_limit:
            params[self.BACKOFF_LIMIT] = self.backoff_limit

        return params


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def checkpoint(self):
        if self.__checkpoint is None:
            return None

        if self.__checkpoint == 'auto':
            return self.checkpoint_table(self.start, self.end)

        return self.__checkpoint


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def topic(self):
        return self.__topic


    @property
    def start(self):
        return self.__start


    @property
    def end(self):
        return self.__end


    @property
    def path(self):
        return self.__path


    @property
    def fetch_last_written(self):
        return self.__fetch_last_written


    @property
    def include_wrapper(self):
        return self.__include_wrapper


    @property
    def rec_only(self):
        return self.__rec_only


    @property
    def min_max(self):
        return self.__min_max


    @property
    def exclude_remainder(self):
        return self.__exclude_remainder


    @property
    def fetch_last_written_before(self):
        return self.__fetch_last_written_before


    @property
    def backoff_limit(self):
        return self.__backoff_limit


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MessageRequest:{topic:%s, start:%s, end:%s, path:%s, fetch_last_written:%s, checkpoint:%s, " \
               "include_wrapper:%s, rec_only:%s, min_max:%s, exclude_remainder:%s, " \
               "fetch_last_written_before:%s, backoff_limit:%s}" % \
               (self.topic, self.start, self.end, self.path, self.fetch_last_written, self.__checkpoint,
                self.include_wrapper, self.rec_only, self.min_max, self.exclude_remainder,
                self.fetch_last_written_before, self.backoff_limit)


# --------------------------------------------------------------------------------------------------------------------

class MessageResponse(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        code = jdict.get('statusCode')
        status = jdict.get('status')
        fetched_last = jdict.get('fetchedLastWrittenData')
        interval = jdict.get('interval')

        items = []
        for msg_jdict in jdict.get('Items'):
            item = Message.construct_from_jdict(msg_jdict) if 'payload' in msg_jdict else msg_jdict
            items.append(item)

        next_url = jdict.get('next')

        return cls(code, status, fetched_last, interval, items, next_url)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, code, status, fetched_last, interval, items, next_url):
        """
        Constructor
        """
        self.__code = int(code)                         # int
        self.__status = status                          # string
        self.__fetched_last = fetched_last              # "Fetched last written data" flag
        self.__interval = interval                      # int

        self.__items = items                            # list of Message

        self.__next_url = next_url                      # URL string


    def __len__(self):
        return len(self.items)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.code is not None:
            jdict['statusCode'] = self.code

        if self.status is not None:
            jdict['status'] = self.status

        if self.fetched_last is not None:
            jdict['fetchedLastWrittenData'] = self.fetched_last

        if self.items is not None:
            jdict['interval'] = None if self.interval is None else int(round(self.interval))

            jdict['Items'] = self.items
            jdict['itemCount'] = len(self.items)

        if self.next_url is not None:
            jdict['next'] = self.next_url

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    def start(self):
        if not self.items:
            return None

        item = self.items[0]

        try:
            return item['rec']
        except TypeError:
            return item.payload['rec']


    def end(self):
        if not self.items:
            return None

        item = self.items[-1]

        try:
            return item['rec']
        except TypeError:
            return item.payload['rec']


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def code(self):
        return self.__code


    @property
    def status(self):
        return self.__status


    @property
    def fetched_last(self):
        return self.__fetched_last


    @property
    def interval(self):
        return self.__interval


    @property
    def items(self):
        return self.__items


    @property
    def next_url(self):
        return self.__next_url


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MessageResponse:{code:%s, status:%s, fetched_last:%s, interval:%s, items:%s, next_url:%s}" % \
               (self.code, self.status, self.fetched_last, self.interval, Str.collection(self.items), self.next_url)



"""
Created on 17 Jan 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from http import HTTPStatus

from scs_core.aws.security.organisation import Organisation, OrganisationPathRoot, OrganisationUser, \
    OrganisationUserPath, OrganisationDevice

from scs_core.data.json import JSONify

from scs_core.sys.http_exception import HTTPException
from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class OrganisationManager(object):
    """
    classdocs
    """

    __MANAGER_URL = "https://04h65m94o8.execute-api.us-west-2.amazonaws.com/default/OrganisationManager"
    __EXECUTIVE_URL = "https://19tm2j6odj.execute-api.us-west-2.amazonaws.com/default/OrganisationExecutive"

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, http_client):
        self.__http_client = http_client                # requests package
        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------
    # Organisation...

    def find_organisations(self, token):
        url = '/'.join((self.__MANAGER_URL, 'organisation'))

        response = self.__http_client.get(url, headers=self.__headers(token))
        self.__check_response(response)

        return (Organisation.construct_from_jdict(jdict) for jdict in response.json())


    def get_organisation_by_label(self, token, label):
        url = '/'.join((self.__MANAGER_URL, 'organisation'))
        payload = JSONify.dumps({"Label": label})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return Organisation.construct_from_jdict(response.json())


    def insert_organisation(self, token, organisation):
        url = '/'.join((self.__EXECUTIVE_URL, 'organisation'))
        payload = JSONify.dumps(organisation)

        response = self.__http_client.post(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return Organisation.construct_from_jdict(response.json())


    def update_organisation(self, token, organisation):
        url = '/'.join((self.__MANAGER_URL, 'organisation'))
        payload = JSONify.dumps(organisation)

        response = self.__http_client.patch(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)


    def delete_organisation(self, token, org_id):
        url = '/'.join((self.__EXECUTIVE_URL, 'organisation'))
        payload = JSONify.dumps({"OrgID": org_id})

        response = self.__http_client.delete(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)


    # ----------------------------------------------------------------------------------------------------------------
    # OrganisationPathRoot...

    def find_oprs_by_organisation(self, token, org_id):
        url = '/'.join((self.__MANAGER_URL, 'opr'))
        payload = JSONify.dumps({"OrgID": org_id})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return (OrganisationPathRoot.construct_from_jdict(jdict) for jdict in response.json())


    def get_opr_by_path_root(self, token, path_root):
        url = '/'.join((self.__MANAGER_URL, 'opr'))
        payload = JSONify.dumps({"PathRoot": path_root})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return OrganisationPathRoot.construct_from_jdict(response.json())


    def insert_opr(self, token, opr):
        url = '/'.join((self.__EXECUTIVE_URL, 'opr'))
        payload = JSONify.dumps(opr)

        response = self.__http_client.post(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return OrganisationPathRoot.construct_from_jdict(response.json())


    def delete_opr(self, token, opr_id):
        url = '/'.join((self.__EXECUTIVE_URL, 'opr'))
        payload = JSONify.dumps({"OPRID": opr_id})

        response = self.__http_client.delete(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)


    # ----------------------------------------------------------------------------------------------------------------
    # OrganisationUser...

    def find_users_by_organisation(self, token, org_id):
        url = '/'.join((self.__MANAGER_URL, 'user'))
        payload = JSONify.dumps({"OrgID": org_id})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return (OrganisationUser.construct_from_jdict(jdict) for jdict in response.json())


    def find_users_by_username(self, token, username):
        url = '/'.join((self.__MANAGER_URL, 'user'))
        payload = JSONify.dumps({"Username": username})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return (OrganisationUser.construct_from_jdict(jdict) for jdict in response.json())


    def get_user(self, token, username, org_id):
        url = '/'.join((self.__MANAGER_URL, 'user'))
        payload = JSONify.dumps({"Username": username, "OrgID": org_id})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return OrganisationUser.construct_from_jdict(response.json())


    def assert_user(self, token, user):
        url = '/'.join((self.__MANAGER_URL, 'user'))
        payload = JSONify.dumps(user)

        response = self.__http_client.post(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)


    def delete_user(self, token, username, org_id):
        url = '/'.join((self.__EXECUTIVE_URL, 'user'))
        payload = JSONify.dumps({"Username": username, "OrgID": org_id})

        response = self.__http_client.delete(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)


    # ----------------------------------------------------------------------------------------------------------------
    # OrganisationUserPath...

    def find_oups(self, token, username, opr_id):
        url = '/'.join((self.__MANAGER_URL, 'oup'))
        payload = JSONify.dumps({"Username": username, "OPRID": opr_id})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return (OrganisationUserPath.construct_from_jdict(jdict) for jdict in response.json())


    def assert_oup(self, token, oup):
        url = '/'.join((self.__MANAGER_URL, 'oup'))
        payload = JSONify.dumps(oup)

        response = self.__http_client.post(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)


    def delete_oup(self, token, oup):
        url = '/'.join((self.__MANAGER_URL, 'oup'))
        payload = JSONify.dumps(oup)

        response = self.__http_client.delete(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)


    # ----------------------------------------------------------------------------------------------------------------
    # OrganisationDevice...

    def find_devices_by_tag(self, token, device_tag):
        url = '/'.join((self.__MANAGER_URL, 'device'))
        payload = JSONify.dumps({"DeviceTag": device_tag})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return (OrganisationDevice.construct_from_jdict(jdict) for jdict in response.json())


    def find_devices_by_organisation(self, token, org_id):
        url = '/'.join((self.__MANAGER_URL, 'device'))
        payload = JSONify.dumps({"OrgID": org_id})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return (OrganisationDevice.construct_from_jdict(jdict) for jdict in response.json())


    def get_device(self, token, device_tag, org_id, device_path, environment_path):
        url = '/'.join((self.__MANAGER_URL, 'device'))
        payload = JSONify.dumps({"DeviceTag": device_tag, "OrgID": org_id, "DevicePath": device_path,
                                 "EnvironmentPath": environment_path})

        response = self.__http_client.get(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)

        return OrganisationDevice.construct_from_jdict(response.json())


    def assert_device(self, token, device):
        url = '/'.join((self.__EXECUTIVE_URL, 'device'))
        payload = JSONify.dumps(device)

        response = self.__http_client.post(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)


    def delete_device(self, token, device_tag, org_id, device_path, environment_path):
        url = '/'.join((self.__EXECUTIVE_URL, 'organisation'))
        payload = JSONify.dumps({"DeviceTag": device_tag, "OrgID": org_id, "DevicePath": device_path,
                                 "EnvironmentPath": environment_path})

        response = self.__http_client.delete(url, data=payload, headers=self.__headers(token))
        self.__check_response(response)


    # ----------------------------------------------------------------------------------------------------------------

    def __headers(self, token):
        headers = {"Content-type": "application/x-www-form-urlencoded", "Accept": "text/json", "Token": token}
        self.__logger.debug('headers: %s' % headers)

        return headers


    def __check_response(self, response):
        self.__logger.debug('response: %s' % response.json())

        status = HTTPStatus(response.status_code)

        if status != HTTPStatus.OK:
            raise HTTPException.construct(status.value, response.reason, response.json())


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "OrganisationManager:{http_client:%s}" % self.__http_client
